package main

import (
	"bufio"
	"errors"
	"fmt"
	"os"
	"strings"
)

func main() {
	stack := Stack{values: []int{}}
	rstack := Stack{values: []int{}}

	dict := Dictionary{}
	b := &Builtins{}
	dict.AddWord("+", b.Add(&stack), nil, false)
	dict.AddWord("-", b.Sub(&stack), nil, false)
	dict.AddWord("*", b.Mul(&stack), nil, false)
	dict.AddWord("/", b.Div(&stack), nil, false)
	dict.AddWord(".", b.Print(&stack), nil, false)
	dict.AddWord("=", b.Eq(&stack), nil, false)
	dict.AddWord("0=", nil, []string{"0", "="}, false)
	dict.AddWord("<>", b.NEq(&stack), nil, false)
	dict.AddWord(">", b.Gt(&stack), nil, false)
	dict.AddWord("<", b.Lt(&stack), nil, false)
	dict.AddWord(">=", b.GtEq(&stack), nil, false)
	dict.AddWord("<=", b.LtEq(&stack), nil, false)
	dict.AddWord("0<", nil, []string{"0", "<"}, false)
	dict.AddWord("0>", nil, []string{"0", ">"}, false)
	dict.AddWord("DUP", b.Dup(&stack), nil, false)
	dict.AddWord("SWAP", b.Swap(&stack), nil, false)
	dict.AddWord("OVER", b.Over(&stack), nil, false)
	dict.AddWord("DROP", b.Drop(&stack), nil, false)
	dict.AddWord("ROT", b.Rot(&stack), nil, false)
	dict.AddWord("WORDS", b.Words(dict), nil, false)
	dict.AddWord(".S", b.Debug(&stack), nil, false)
	dict.AddWord(".R", b.Debug(&rstack), nil, false)
	dict.AddWord("EMIT", b.Emit(&stack), nil, false)
	dict.AddWord("R>", b.RFrom(&stack, &rstack), nil, false)
	dict.AddWord(">R", b.ToR(&stack, &rstack), nil, false)
	dict.AddWord("R@", b.RFetch(&stack, &rstack), nil, false)
	dict.AddWord("DO", b.Do(&stack, &rstack), nil, false)
	dict.AddWord("LOOP", b.Loop(&stack, &rstack), nil, false)
	dict.AddWord("I", b.I(&stack, &rstack), nil, false)
	dict.AddWord("CR", b.CR(), nil, false)
	dict.AddWord("QUIT", b.Quit(), nil, false)

	c := Context{
		Dictionary: dict,
		Stack:      &stack,
		RStack:     &rstack,
	}

	reader := bufio.NewReader(os.Stdin)
	fmt.Print("prosper")

	// read loop
	for {
		fmt.Print("\n> ")

		line, err := reader.ReadString('\n')
		if err != nil {
			fmt.Println(err.Error())
			os.Exit(1)
		}
		line = strings.TrimSpace(line)
		err = c.Eval(line + " ") // append a space to make sure we always close out our parse loop
		if errors.Is(err, ErrExit) {
			fmt.Printf("bye\n")
			os.Exit(0)
		} else if err != nil {
			fmt.Printf("error in evaluation: %v", err)
		} else {
			fmt.Print("ok")
		}
	}
}
